#ifndef GUI_H
#define GUI_H

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

typedef long guiNativeHandle;

/* All the low-level canvas stuff. */ 

typedef struct {
	int x;
	int y;
} guiPoint, *guiPointPtr;

typedef struct {
	int w;
	int h;
} guiSize;

typedef struct {
	int x;
	int y;
	int w;
	int h;
} guiRectangle, *guiRectanglePtr; 

extern guiRectangle guiRectangleEmpty;

/* 0-255. Declared as int for efficiency and ease 
	of adjusting for overflow, dithering, etc. */

typedef struct {
	int r;
	int g;
	int b;
} guiColor, *guiColorPtr;

typedef enum {
	guiCourier,
	guiTimesRoman,
	guiHelvetica,
	guiFontFamiliesTotal
} guiFontFamily;

typedef struct {
	guiFontFamily family;
	int size;
	int italic;
	int bold;
	int ascent;
	int descent;
} guiFont, *guiFontPtr;

struct guiEventStruct;

typedef enum {
	guiCanvasApplication,
	guiCanvasDialog,
	guiCanvasUndecorated
} guiCanvasType;

typedef enum {
	guiBorderBackground,
	guiBorderIn,
	guiBorderOut
} guiBorderStyle;

typedef struct guiCanvasStruct {
	/* DON'T manipulate or even read
		this stuff -- use the provided functions. */
	struct guiCanvasStruct *parent;
	/* Next sibling */
	struct guiCanvasStruct *next;
	/* First child */
	struct guiCanvasStruct *first;
	/* Child with focus (for tab cycling) */
	struct guiCanvasStruct *childFocused;
	guiPoint l;
	guiSize physicalSize;
	guiSize virtualSize;
	guiPoint visibleTop;
	int barV;
	int barH;
	int repaintV;
	int repaintH;
	int repaintB;
	int border;
	int noScrollV;
	int noScrollH;
	guiBorderStyle borderStyle;
	guiRectangle invalid;
	guiNativeHandle handle;
	guiNativeHandle drawContext;
	guiNativeHandle fillContext;
	int offscreen;
	int (*eventHandler) (struct guiCanvasStruct *c, 
		struct guiEventStruct *e);
	int visible;
	int repainting;
	guiColor drawColor;
	guiColor textBackgroundColor;
	guiColor fillColor;
	guiColor backgroundColor;
	int dim;
	int xorFlag;
	int lineWidth;
	int dying;
	guiFont font;
	FILE *psoutput;
	int pslastfill;
	void *clientData;
	int eventMask;
	int virtualized;
	int fontSet;
	int clearing;
	guiCanvasType t;
	struct guiCanvasStruct *onBehalfOf;
	struct guiCanvasStruct *icon;
	int wantFocus;
	/* Has keyboard focus (this canvas specifically,
		not one of its children) */
	int focused;
	int tabbing;
} guiCanvas, *guiCanvasPtr;

typedef enum {
	guiBitmapJpeg,
	guiBitmapPng,
	guiBitmapGif
} guiBitmapSourceType;

typedef struct {
	guiSize s;
	guiBitmapSourceType sourceType;
	guiNativeHandle handle;
} guiBitmap, *guiBitmapPtr;

/* EVENTS */

typedef enum {
        guiEventRepaint = 1,
        guiEventKeyDown = 2,
        guiEventResize = 8,
        guiEventAction = 16,
        guiEventMouseDown = 32,
        guiEventMouseUp = 64,
        guiEventMouseMove = 128,
        guiEventMouseDrag = 256,
        guiEventMouseEnter = 512,
        guiEventMouseExit = 1024,
        guiEventTimer = 2048,
	guiEventDestroyed = 4096,
	guiEventFocusIn = 8192,
	guiEventFocusOut = 16384
} guiEventType;

typedef enum {
	guiKeyLeft = -1,
	guiKeyRight = -2,
	guiKeyDown = -3,
	guiKeyUp = -4,
	guiKeyInsert = -5,
	guiKeyDelete = -6,
	guiKeyHome = -7,
	guiKeyEnd = -8,
	guiKeyPageUp = -9,
	guiKeyPageDown = -10,
	guiKeyF1 = -12,
	guiKeyF2 = -13,
	guiKeyF3 = -14,
	guiKeyF4 = -15,
	guiKeyF5 = -16,
	guiKeyF6 = -17,
	guiKeyF7 = -18,
	guiKeyF8 = -19,
	guiKeyF9 = -20,
	guiKeyF10 = -21,
	guiKeyF11 = -22,
	guiKeyF12 = -23,
	guiKeyNone = -24,
	guiKeyBackspace = 8,
	guiKeyEscape = 27,
	guiKeyTab = 9,
	guiKeyEnter = 13,
	guiKeySpace = 32
	/* ASCII keys have their normal ASCII values */
} guiKey;

typedef enum {
	guiButtonLeft = 1,
	guiButtonMiddle = 2,
	guiButtonRight = 4
} guiButtonType;

/* Fields not relevant to a particular event
	have undefined values! */

typedef struct guiEventStruct {
        guiEventType t;
        guiPoint l;
        guiSize s;
        guiKey key;
	guiRectangle area;
	int shift;
	int alt;
	int control;
	guiButtonType button;
        unsigned char ascii;
        int timerid;
} guiEvent, *guiEventPtr;

#define guiNone (-1)

/* Called first. */

int guiSetup();

/* Called last, after guiEventLoop() returns. */

void guiShutdown();

/* Called after guiSetup() and the initial creation of various objects.
	Does not return until the application has been exited. */

void guiMainLoop();

guiSize guiGetScreenSize();

guiPoint guiGetMouseLocation();

/* Creates a canvas. */

guiCanvasPtr guiCanvasCreate(guiCanvasPtr parent, int x, int y, int w, int h);

guiCanvasPtr guiCanvasCreateOffscreen(int w, int h);

/* Names a canvas. This is only relevant for top-level canvases (no parent). */
void guiCanvasSetName(guiCanvasPtr c, char *name);

/* Sets the icon for a canvas. Again, only relevant for top-level canvases. */

void guiCanvasSetIcon(guiCanvasPtr c, guiBitmapPtr b);

/* Border pixels (per edge) and the style to draw the border in. 
	Default is no border area. */

void guiCanvasSetBorder(guiCanvasPtr c, int b, guiBorderStyle style);

/* This is normally used to indicate that a top-level canvas
	should have less of, or none of, the usual window dressing 
	the operating system would add. This call is completely ignored
	for child canvases. Useful for implementing pop-up menus, 
	dialog boxes and the like. The third parameter is used to 
	indicate the application canvas that a dialog box or menu
	is being popped up in regard to, if any, and should be null for 
	normal guiCanvasApplication canvases. */
 
void guiCanvasSetType(guiCanvasPtr c, guiCanvasType t,
	guiCanvasPtr onBehalfOf);

void guiCanvasDestroy(guiCanvasPtr c);

/* Displays the dialog in a reasonable position for a dialog box,
	then invokes guiCanvasModal(). */
	
void guiCanvasDoDialog(guiCanvasPtr c);

/* Same, but doesn't invoke guiCanvasModal(). */
void guiCanvasDoModelessDialog(guiCanvasPtr c);

/* This function does not return until the specified canvas
	has been hidden or destroyed. Child canvases will
	receive events normally, but unrelated canvases
	will not receive user input events (they will receive
	repaint and resize events). This is used to 
	implement modal dialog boxes and the like. 
	Calls to this function can be nested for different
	canvases. */

void guiCanvasModal(guiCanvasPtr c);

void guiCanvasSetBackgroundColor(guiCanvasPtr c, guiColor b);

/* Used to store and retrieve a pointer to a higher-level
	object that owns this canvas; often needed
	in the event handler. */
void guiCanvasSetClientData(guiCanvasPtr c, void *clientData);

void *guiCanvasGetClientData(guiCanvasPtr c);

/* Event function should return true if the event has been 
	satisfactorily handled and should not be considered further. 
	This can be used to inhibit default behaviors, such as
	automatic arrow key support in scrollbars, etc. */

void guiCanvasSetEventHandler(guiCanvasPtr c,
	int (*eventHandler) (guiCanvasPtr c, guiEventPtr e));

/* Bitwise OR */
void guiCanvasSetEventMask(guiCanvasPtr c, guiEventType t);

/* Returns true if the timer was set successfully.
	Timer ids must be positive integers. Otherwise
	guiCanvasSetTimer() will return false. */

int guiCanvasSetTimer(guiCanvasPtr c, int id, long milliseconds);

void guiCanvasClearTimer(guiCanvasPtr c, int id);

void guiCanvasMove(guiCanvasPtr c, int x, int y);

/* Sets both physical and virtual size (they will be the same). */
void guiCanvasSetSize(guiCanvasPtr c, int w, int h);

/* Sets only the virtual size. Scrollbars will appear if
	necessary to allow navigation. */
void guiCanvasSetVirtualSize(guiCanvasPtr c, int w, int h);

/* Sets only the physical size. Scrollbars will appear if
	necessary to allow navigation. */
void guiCanvasSetPhysicalSize(guiCanvasPtr c, int w, int h);

/* Returns the current visible area. This will be the same
	as the physical size, except when scrollbars are
	present, taking up some of the space. */
guiSize guiCanvasGetVisibleSize(guiCanvasPtr c);

/* If true, the canvas will be mapped, otherwise hidden. */
void guiCanvasShow(guiCanvasPtr c, int show);

/* Returns true if this canvas is an offscreen drawing area. */
int guiCanvasIsOffscreen(guiCanvasPtr c);

/* Returns true if this canvas and all of its ancestry are visible
	(ie, it is truly visible to the user!). "Offscreen"
	drawing canvases are never visible. */

int guiCanvasIsVisible(guiCanvasPtr c);

/* Returns true if this canvas is currently being rendered to a printer. */
int guiCanvasIsPrinting(guiCanvasPtr c);

guiPoint guiCanvasGetPosition(guiCanvasPtr c);

guiPoint guiCanvasGetScreenPosition(guiCanvasPtr c);

guiSize guiCanvasGetVirtualSize(guiCanvasPtr c);

guiSize guiCanvasGetPhysicalSize(guiCanvasPtr c);

guiCanvasPtr guiCanvasGetParent(guiCanvasPtr c);

/* Returns the top-level parent of the canvas. Might
	be the same canvas, of course. */
guiCanvasPtr guiCanvasGetRootParent(guiCanvasPtr c);

/* If true (the default), areas that need to be repainted
	are automatically erased (to the background color)
	before the repaint event is set. If false, the
	application is responsible for repainting
	every pixel. */

void guiCanvasSetClearing(guiCanvasPtr c, int flag);

/* Indicates whether this canvas is interested
	in the keyboard focus. If set, the parent
	canvas will allow the user to tab to
	this canvas. */

void guiCanvasWantFocus(guiCanvasPtr c, int flag);

/* Requests that the keyboard focus be given to this
	particular canvas immediately. */

void guiCanvasTakeFocus(guiCanvasPtr c);

/* Returns whether this canvas has the keyboard focus. */

int guiCanvasQueryFocus(guiCanvasPtr c);

/* Turns on or off tab-cycling support in this canvas,
	which is presumably the parent of several 
	input controls. */

void guiCanvasSetTabbing(guiCanvasPtr c, int flag);

/* Size of a printer page */
guiSize guiGetPageSize();

void guiCanvasInvalidate(guiCanvasPtr c, guiRectanglePtr r);

void guiCanvasSetDrawColor(guiCanvasPtr c, guiColor color);

/* Only when guiCanvasDrawOpaqueText is used */
void guiCanvasSetTextBackgroundColor(guiCanvasPtr c, guiColor color);

void guiCanvasSetFillColor(guiCanvasPtr c, guiColor color);

void guiCanvasDrawLine(guiCanvasPtr c, int x1, int y1, int x2, int y2);

void guiCanvasDrawPoint(guiCanvasPtr c, guiPoint p);

/* "Dims" draws and fills, typically by using
	dotted lines and pattern fills. This is
	useful to draw "inactive" versions of controls. */
void guiCanvasSetDim(guiCanvasPtr c, int flag);

void guiCanvasSetXOR(guiCanvasPtr c, int xor);

void guiCanvasSetLineWidth(guiCanvasPtr c, int width);

void guiCanvasSetFont(guiCanvasPtr c, guiFontPtr f);

void guiCanvasDrawPolygon(guiCanvasPtr c, 
	guiPoint *points, int pointsTotal);

void guiCanvasDrawPolyline(guiCanvasPtr c, 
	guiPoint *points, int pointsTotal);

void guiCanvasFillPolygon(guiCanvasPtr c, 
	guiPoint *points, int pointsTotal);

void guiCanvasDrawRectangle(guiCanvasPtr c, int x, int y, int w, int h);

void guiCanvasFillRectangle(guiCanvasPtr c, int x, int y, int w, int h); 

/* An ellipse with its center at x and y */
void guiCanvasDrawEllipse(guiCanvasPtr c, int x, int y, int w, int h); 

void guiCanvasFillEllipse(guiCanvasPtr c, int x, int y, int w, int h); 

void guiCanvasDrawText(guiCanvasPtr c, int x, int y, char *text, int len);

void guiCanvasDrawOpaqueText(guiCanvasPtr c, int x, int y, char *text, int len);

void guiCanvasCopyArea(guiCanvasPtr dst, guiCanvasPtr src,
	int dstx, int dsty, int srcx, int srcy, int w, int h);

void guiCanvasDrawBitmap(guiCanvasPtr c, int x, int y, guiBitmapPtr b);

void guiCanvasDrawBitmapSection(guiCanvasPtr c, 
	int dstx, int dsty, 
	int srcx, int srcy, int w, int h,
	guiBitmapPtr b);

guiBitmapPtr guiCanvasGrabBitmap(guiCanvasPtr c, 
	int x, int y, int w, int h);

/* Passes all matching keyboard events to this canvas. 
	When grabs conflict, grabs by children of the
	same top-level parent take precedence. 

	For normal ASCII characters, the value of
	'shift' will be IGNORED. Set guiKey
	to ascii A (65) if you want to grab uppercase A,
	and don't worry about the value of shift. */

void guiCanvasGrabKey(guiCanvasPtr c, guiKey k, 
	int shift, int control, int alt);

/* Releases a keystroke grab. */

void guiCanvasReleaseKey(guiCanvasPtr c, guiKey k,
	int shift, int control, int alt);

/* Print to the specified output file in Postscript format.
	Prints the virtual canvas (entire on the Y axis,
	as multiple pages broken at guiGetPageSize().h 
	intervals; clipped to guiGetPageSize().w on the X axis). */

void guiCanvasPrint(guiCanvasPtr c, FILE *out);

/* Shrinks to the closest font that can actually be rendered. */
void guiFontFindClosest(guiFontPtr f);

/* Determines the size of text in the specified font. */
guiSize guiFontTextSize(guiFontPtr f, char *text, int len);

/* Bitmaps are created by guiCanvasGrabBitmap (above)
	or guiBitmapLoad or guiBitmapPortion (below). */

/* Loads a bitmap from a file. This stream must be rewindable
	(not stdin). If the dither flag is false, the closest
	pure color available will be used instead for each pixel.
	This is unadvisable except for simple icons. */

guiBitmapPtr guiBitmapLoad(FILE *in, int dither);

/* Loads a bitmap from a buffer of unsigned char data in memory.
	This bitmap must be in PNG format. See the
	guires utility for a way of converting a PNG
	image file to a C source file defining such
	a buffer, which can be linked to your program. 
	See also guiBitmapRaw. If the dither flag is false, the closest
	pure color available will be used instead for each pixel.
	This is unadvisable except for simple icons. */

guiBitmapPtr guiBitmapBuiltin(unsigned char *resource, int dither);

/* Loads a bitmap from a buffer of raw pixel values. Specifically,
	each pixel is represented by one byte for red,
	one byte for green, and one byte for blue, where
	0 is black and 255 is maximum intensity. 
	The top row of pixels should appear first
	in the buffer, from left to right. If the dither flag is 
	false, the closest pure color available will be used instead 
	for each pixel.  This is unadvisable except for simple icons. */

guiBitmapPtr guiBitmapRaw(unsigned char *resource, int w, int h, int dither);

/* Copies a bitmap. */

guiBitmapPtr guiBitmapCopy(guiBitmapPtr b);

/* Returns a scaled version of a portion of a bitmap. 
	The original bitmap is not changed. */

guiBitmapPtr guiBitmapPortion(guiBitmapPtr src,
	int srcx, int srcy,
	int srcw, int srch,
	int dstw, int dsth);

/* Saves a bitmap to a file. This is not lossless -- the bitmap's
	format will often have been changed to be appropriate for 
	the display. You must specify the format to save in;
	b->sourceType is usually a good value, but I require
	this argument because most programmers will forget
	that the default sourceType for a grabbed bitmap is PNG. 

	Returns true on a successful save, false on I/O errors. 

	WARNING WARNING WARNING

	Bitmaps get dithered appropriately for the native display
	when they are loaded. This means that these functions are
	probably not well suited for a lossless image editor
	(saving a new version will modify the pixels). */

int guiBitmapSave(guiBitmapPtr b, FILE *out, guiBitmapSourceType sourceType);

/* Destroys a bitmap. */
void guiBitmapDestroy(guiBitmapPtr b);

/* Displays a message and waits for the user's OK. Returns
	one of the first four values below. mtype can be
	guiMessageOK or one of the last three values below. */

enum {
	guiMessageOK,
	guiMessageYes,
	guiMessageNo,
	guiMessageCancel,
	guiMessageYesNo,
	guiMessageOKCancel,
	guiMessageYesNoCancel
};

int guiMessage(guiCanvasPtr parent, char *title, char *text, int mtype);

/* Displays an error message, waits for the user's OK, 
	and exits the program. Arguments are the same
	as for printf(). This is intended for
	catastrophic failures only. */
void guiFatalError(char *format, ...);

/* Scrollbar size constants. */

#define guiVerticalScrollbarWidth 16
#define guiHorizontalScrollbarHeight 16
#define guiScrollbarStepPixels 8

extern guiColor guiDefaultBackgroundColor;

#define guiMax(a, b) ((a) > (b) ? (a) : (b))
#define guiMin(a, b) ((a) < (b) ? (a) : (b))
#define guiBound(x, low, high) (((x) < (low)) ? (low) : \
        ((x) > (high)) ? (high) : (x))

/* Returns true if the specified point in the virtual document
	is currently visible in the physical window. Often used
	to test whether a call to guiCanvasScrollTo() is in order. */

int guiCanvasPointIsVisible(guiCanvasPtr c, int x, int y);

/* Scrolls to make the entire rectangle specified
	visible, or the upper left corner if that
	is impossible. */

void guiCanvasScrollToRectangle(guiCanvasPtr c, guiRectanglePtr r);

void guiCanvasScrollToVertical(guiCanvasPtr c, int pos);

void guiCanvasScrollToHorizontal(guiCanvasPtr c, int pos);

void guiCanvasScrollTo(guiCanvasPtr c, int x, int y);

void guiCanvasScrollVertical(guiCanvasPtr c, int delta);

void guiCanvasScrollHorizontal(guiCanvasPtr c, int delta);

void guiCanvasScroll(guiCanvasPtr c, int dx, int dy);

/* Indicates whether a vertical scrollbar should appear (when needed). */
void guiCanvasAllowVerticalScrollbar(guiCanvasPtr c, int flag);

/* Indicates whether a horizontal scrollbar should appear (when needed). */
void guiCanvasAllowHorizontalScrollbar(guiCanvasPtr c, int flag);

/* Returns the clipboard contents as a null-terminated string;
	also sets *pLength to the length (not including the null)
	if pLength is not null. The returned value must be
	freed with free() after use! */

char *guiGetClipboard(int *pLength);

/* Sets the clipboard to the data specified. */

void guiSetClipboard(char *data, int length);

/* Tests whether a key represents a printable character. */

int guiKeyIsPrintable(guiKey k);

#endif /* GUI_H */

